#ifndef ARITH_H
#define ARITH_H

//
// The DCC95 arithmetic coder
//

#include "types.h"
#include "macros.h"
#include "histo.h"
#include "iobuf.h"

class Arith {
public:
                 Arith(InputBuf *iBuf, OutputBuf *oBuf, int bits);
    void         init(InputBuf *iBuf, OutputBuf *oBuf, int bits);
    void         startEncoding();
    void         finishEncoding();
    void         startDecoding();
    void         encodeBlock(Histogram *H);
    void         decodeBlock(Histogram *H);
    inline void  encodeSymbol(int symbol, Histogram *H);
    inline int   decodeSymbol(Histogram *H);
protected:
    inline void  putBit(int bit)
    {
        oBuf->putBit(bit);
        while (followBits > 0)
        {
            oBuf->putBit(!bit);
            followBits--;
        }
    }
    int          bits;
    word         half, qtr;
    word         L, R, followBits;
    InputBuf    *iBuf;
    OutputBuf   *oBuf;
};


inline void Arith::encodeSymbol(int symbol, Histogram *H)
{
    word l = 0,
         h,
         t = H->getTotFreq(),
         r = R / t,
         rl;

    for (int i = 0; i < symbol; i++)
    {
        l += H->getFreq(i);
    }
    h = l + H->getFreq(symbol);
    rl = r * l;
    L += rl;
    if (h < t)
    {
        R = r * h - rl;
    }
    else
    {
        R -= rl;
    }
    while (R <= qtr)
    {
        if (L + R <= half)
        {
            putBit(0);
        }
        else if (L >= half)
        {
            putBit(1);
            L -= half;
        }
        else
        {
            followBits++;
            L -= qtr;
        }
        L <<= 1;
        R <<= 1;
    }
    H->update(symbol);
}


int Arith::decodeSymbol(Histogram *H) {
    word l,
         h = 0,
         t = H->getTotFreq(),
         r = R / t, rl,
         target = min(t - 1, L / r);
    int  bit,
         symbol = -1;

    while (h <= target)
    {
        h += H->getFreq(++symbol);
    }
    l = h - H->getFreq(symbol);
    rl = r * l;
    L -= rl;
    if (h < t)
    {
        R = r * h - rl;
    }
    else
    {
        R -= rl;
    }
    while (R <= qtr)
    {
        R <<= 1;
        bit = iBuf->getBit();
        if (bit == -1)
        {
            return -1;
        }
        L = (L << 1) | bit;
    }
    H->update(symbol);
    return symbol;
}

#endif
