#include <assert.h>
#include "types.h"
#include "iobuf.h"

InputBuf::InputBuf(FILE *f, void *buf, int len) {
  init(f, buf, len);
}

InputBuf::InputBuf(void *buf, int len) {
  init(NULL, buf, len);
}

void
InputBuf::init(FILE *f, void *buf, int len) {
  InputBuf::f = f;
  InputBuf::buf = (byte*)buf;
  InputBuf::len = len;
  charsRead = 0;
  if (f) refresh();
  else {
    p = (byte*)buf;
    pLimit = p + len;
    iMask = 0x80;
  }
}

int InputBuf::readWord(word *w) {
  int   i, c;
  byte  buf[sizeof(word)];
  word  res = 0;

  if (iMask != 0x80) {
    if (getChar() < 0) return -1;
    iMask = 0x80;
  }
  for (i = 0; i < (int)sizeof(word); i++) {
    c = getChar(); if (c < 0) return -1;
    buf[i] = c;
  }
  res = (buf[0] << 24) |
        (buf[1] << 16) |
        (buf[2] << 8) |
        (buf[3]);
  *w = res;
  return 0;
}

void
InputBuf::refresh() {
  int    n;

  p = buf;
  iMask = 0x80;
  n = fread(p, 1, len, f);
  charsRead += n;
  pLimit = p + n;
}

OutputBuf::OutputBuf(FILE *f, void *buf, int len) {
  init(f, buf, len);
}

OutputBuf::OutputBuf(void *buf, int len) {
  init(NULL, buf, len);
}

OutputBuf::~OutputBuf() {
  if (f) flush();
}

void
OutputBuf::init(FILE *f, void *buf, int len) {
  OutputBuf::f = f;
  OutputBuf::buf = (byte*)buf;
  OutputBuf::len = len;
  p = (byte*)buf;
  pLimit = (byte*)buf + len;
  oMask = 0x80;
  *p = 0;
  charsWritten = 0;
}

int OutputBuf::writeWord(word w) {
  if (oMask != 0x80) {
    if (++p == pLimit) {
      if (f) flush();
      else return -1;
    }
  }
  putChar((w & 0xff000000) >> 24);
  putChar((w & 0x00ff0000) >> 16);
  putChar((w & 0x0000ff00) >> 8);
  putChar((w & 0x000000ff));
  return 0;
}

void
OutputBuf::flush() {
  int    n;

  if (oMask != 0x80) p++;
  if (p - buf) {
    n = fwrite(buf, 1, p - buf, f);
    assert(n == (p - buf));
    charsWritten += n;
  }
  p = buf;
  *p = 0;
  oMask = 0x80;
}
