#ifndef IOBUF_H
#define IOBUF_H


#include <stdio.h>
#include "types.h"


class InputBuf
{
public:
                  InputBuf(FILE *f, void *buf, int len);
                  InputBuf(void *buf, int len);
    void          init(FILE *f, void *buf, int len);
    int           readWord(word *w);
    inline int    getBit();
    inline int    peekChar();
    inline int    getChar();
    inline word   nCharsLeft();
    inline word   nCharsRead() { return charsRead; };
    void          refresh();
protected:
    FILE         *f;
    byte         *buf, *p, *pLimit;
    int           len;
    int           iMask;
    word          charsRead;
};


class OutputBuf
{
public:
                  OutputBuf(FILE *f, void *buf, int len);
                  OutputBuf(void *buf, int len);
                 ~OutputBuf();
    void          init(FILE *f, void *buf, int len);
    int           writeWord(word w);
    inline int    putBit(int bit);
    inline int    putChar(int c);
    inline word   nCharsLeft();
    inline word   nCharsWritten() { return charsWritten; };
    void          flush();
protected:
    FILE         *f;
    byte         *buf, *p, *pLimit;
    int           len;
    int           oMask;
    word          charsWritten;
};


int
InputBuf::getBit()
{
    if (p == pLimit)
    {
        if (f != NULL)
        {
            refresh();
        }
        if (p == pLimit)
        {
            return -1;
        }
    }
    int res = ((*p & iMask) != 0) ? 1 : 0;
    iMask >>= 1;
    if (iMask == 0)
    {
        iMask = 0x80;
        p++;
    }
    return res;
}


int
InputBuf::peekChar()
{
    if (p == pLimit)
    {
        if (f != NULL)
        {
            refresh();
        }
        if (p == pLimit)
        {
            return -1;
        }
    }
    return *p;
}


int
InputBuf::getChar()
{
    if (p == pLimit)
    {
        if (f != NULL)
        {
            refresh();
        }
        if (p == pLimit)
        {
            return -1;
        }
    }
    return *p++;
}


word
InputBuf::nCharsLeft()
{
    if ((p == pLimit) && (f != NULL))
    {
        refresh();
    }
    return pLimit - p;
}


int
OutputBuf::putBit(int bit)
{
    if (p == pLimit)
    {
        if (f != NULL)
        {
            flush();
        }
        else
        {
            return -1;
        }
    }
    if (bit != 0)
    {
        *p |= oMask;
    }
    oMask >>= 1;
    if (oMask == 0)
    {
        oMask = 0x80;
        if (++p < pLimit)
        {
            *p = 0;
        }
    }
    return 0;
}


int
OutputBuf::putChar(int c)
{
    if (p == pLimit)
    {
        if (f != NULL)
        {
            flush();
        }
        else return -1;
    }
    *p++ = c;
    return 0;
}


word
OutputBuf::nCharsLeft()
{
    if ((p == pLimit) && (f != NULL))
    {
        flush();
    }
    return len - (p - buf);
};

#endif
