#include "types.h"
#include "arith.h"
#include "bwt.h"
#include "histo.h"
#include "mtf.h"


#define M0_ZRUN1     256
#define M0_ZRUN2     257
#define M0_EOB       258
#define M0_ONE       259
#define M0_M2        260
#define M0_M4        261
#define M0_M8        262
#define M0_M16       263
#define M0_M32       264
#define M0_M64       265
#define M0_M128      266


void                 MTFinit();
void                 MTFdeInit();
void                 MTFstartEncoding();
void                 MTFfinishEncoding();
void                 MTFstartDecoding();
void                 MTFencodeRandomWord();
word                 MTFdecodeRandomWord();
void                 MTFencodeBlock(void);
void                 MTFdecodeBlock(int len);


int                  incVal[8] = { 9, 5, 2, 1, 1, 1, 1, 1 };


static void          MTFsendZeroes();
static inline void   MTFencodeSymbol(int symbol);
static inline int    MTFdecodeSymbol();
static void          MTFencodeRandom(int symbol);
static int           MTFdecodeRandom();


static Arith        *A;
static Histogram    *H[9];
static int           nZeroes;
static int          *pos;


static inline int MTFdecodeSymbol()
{
    switch (A->decodeSymbol(H[0]))
    {
    case -1:               return -1;
    case M0_ZRUN1 - 256:   return M0_ZRUN1;
    case M0_ZRUN2 - 256:   return M0_ZRUN2;
    case M0_EOB - 256:     return M0_EOB;
    case M0_ONE - 256:     return 1;
    case M0_M2 - 256:      return A->decodeSymbol(H[1]) + 2;
    case M0_M4 - 256:      return A->decodeSymbol(H[2]) + 4;
    case M0_M8 - 256:      return A->decodeSymbol(H[3]) + 8;
    case M0_M16 - 256:     return A->decodeSymbol(H[4]) + 16;
    case M0_M32 - 256:     return A->decodeSymbol(H[5]) + 32;
    case M0_M64 - 256:     return A->decodeSymbol(H[6]) + 64;
    default:               return A->decodeSymbol(H[7]) + 128;
    }
}


static inline void MTFencodeSymbol(int symbol)
{
   switch (symbol)
   {
   case M0_ZRUN1:          A->encodeSymbol(M0_ZRUN1 - 256, H[0]); break;
   case M0_ZRUN2:          A->encodeSymbol(M0_ZRUN2 - 256, H[0]); break;
   case M0_EOB:            A->encodeSymbol(M0_EOB - 256, H[0]); break;
   case 1:                 A->encodeSymbol(M0_ONE - 256, H[0]); break;
   case 2:
   case 3:
       A->encodeSymbol(M0_M2 - 256, H[0]);
       A->encodeSymbol(symbol - 2, H[1]);
       break;
   case 4:
   case 5:
   case 6:
   case 7:
       A->encodeSymbol(M0_M4 - 256, H[0]);
       A->encodeSymbol(symbol - 4, H[2]);
       break;
   case 8:
   case 9:
   case 10:
   case 11:
   case 12:
   case 13:
   case 14:
   case 15:
       A->encodeSymbol(M0_M8 - 256, H[0]);
       A->encodeSymbol(symbol - 8, H[3]);
       break;
   default:
       if (symbol >= 16 && symbol <= 31)
       {
           A->encodeSymbol(M0_M16 - 256, H[0]);
           A->encodeSymbol(symbol - 16, H[4]);
       }
       else if (symbol >= 32 && symbol <= 63)
       {
           A->encodeSymbol(M0_M32 - 256, H[0]);
           A->encodeSymbol(symbol - 32, H[5]);
       }
       else if (symbol >= 64 && symbol <= 127)
       {
           A->encodeSymbol(M0_M64 - 256, H[0]);
           A->encodeSymbol(symbol - 64, H[6]);
       }
       else
       {
           A->encodeSymbol(M0_M128 - 256, H[0]);
           A->encodeSymbol(symbol - 128, H[7]);
       }
   }
}


void
MTFinit(InputBuf *iBuf, OutputBuf *oBuf)
{
    A = new Arith(iBuf, oBuf, 26);
    H[0] = new Histogram(11,  incVal[0],  1000);
    H[1] = new Histogram(2,   incVal[1],  1000);
    H[2] = new Histogram(4,   incVal[2],  1000);
    H[3] = new Histogram(8,   incVal[3],  1000);
    H[4] = new Histogram(16,  incVal[4],  1000);
    H[5] = new Histogram(32,  incVal[5],  1000);
    H[6] = new Histogram(64,  incVal[6],  1000);
    H[7] = new Histogram(128, incVal[7],  1000);
    H[8] = new Histogram(256, 0,          1000);
    pos = new int[256];
    for (int i = 0; i < 256; i++)
    {
        pos[i] = i;
    }
    nZeroes = 0;
}


void
MTFdeInit()
{
    delete A;
    for (int i = 0; i <= 8; i++)
    {
        delete H[i];
    }
    delete [] pos;
}


void
MTFstartEncoding()
{
    A->startEncoding();
}


void
MTFfinishEncoding()
{
    if (nZeroes > 0)
    {
        MTFsendZeroes();
    }
    MTFencodeSymbol(M0_EOB);
    A->finishEncoding();
}


void
MTFstartDecoding()
{
    A->startDecoding();
}


void
MTFencodeRandomWord(word w)
{
    MTFencodeRandom((w & 0xff000000) >> 24);
    MTFencodeRandom((w & 0x00ff0000) >> 16);
    MTFencodeRandom((w & 0x0000ff00) >> 8);
    MTFencodeRandom((w & 0x000000ff));
}


word
MTFdecodeRandomWord()
{
    word  b[sizeof(word)];

    for (int i = 0; i < (int)sizeof(word); i++)
    {
        b[i] = MTFdecodeRandom();
    }
    return (b[0] << 24) | (b[1] << 16) | (b[2] << 8) | b[3];
}


void
MTFencodeBlock()
{
    int   c, j, k, t, t2;
    int  *mypos;
    word  i;

    for (i = 0; i < curBlockSize; i++)
    {
        j = ptr[i] - 1;
        if (j < 0)
        {
            j += curBlockSize;
        }
        c = stripe[j] >> 24;
        k = 0;
        mypos = pos;
        t = *mypos;
        while (t != c)
        {
            t2 = t;
            t = *++mypos;
            k++;
            *mypos = t2;
        }
        pos[0] = t;
        if (k > 0)
        {
            if (nZeroes > 0)
            {
                MTFsendZeroes();
            }
            MTFencodeSymbol(k);
        }
        else
        {
            nZeroes++;
        }
    }
}


void
MTFdecodeBlock()
{
    int    c;
    byte  *p = block;
  
    for (;;)
    {
        c = MTFdecodeSymbol();
        if (c < 0)
        {
            return;
        }
        if ((c == M0_ZRUN1) || (c == M0_ZRUN2))
        {
            nZeroes = 0;
            do
            {
                nZeroes <<= 1;
                if (c == M0_ZRUN2)
                {
                    nZeroes |= 1;
                }
                nZeroes++;
                c = MTFdecodeSymbol();
            } while ((c == M0_ZRUN1) || (c == M0_ZRUN2));
            while (nZeroes-- > 0)
            {
                *p++ = pos[0];
            }
        }
        if (c == M0_EOB)
        {
            curBlockSize = p - block;
            return;
        }
        *p++ = pos[c];
        word t = pos[c];
        for (int i = c; i > 0; i--)
        {
            pos[i] = pos[i - 1];
        }
        pos[0] = t;
    }
}


void
MTFencodeRandom(int symbol)
{
    A->encodeSymbol(symbol, H[8]);
}


int
MTFdecodeRandom()
{
    return A->decodeSymbol(H[8]);
}


void
MTFsendZeroes()
{
    int nBits = 0;
    word bits = 0;
    while (nZeroes > 0)
    {
        bits <<= 1;
        nZeroes--;
        if ((nZeroes & 1) != 0)
        {
            bits |= 1;
        }
        nZeroes >>= 1;
        nBits++;
    }
    while (nBits-- > 0)
    {
        if (bits & 1)
        {
            MTFencodeSymbol(M0_ZRUN2);
        }
        else
        {
            MTFencodeSymbol(M0_ZRUN1);
        }
        bits >>= 1;
    }
}
