#include "types.h"
#include "rle.h"


#define MAX_RUN_LENGTH  0xff


RLE::RLE(InputBuf *iBuf, OutputBuf *oBuf, ushort minRunLength)
{
    init(iBuf, oBuf, minRunLength);
}


void
RLE::init(InputBuf *iBuf, OutputBuf *oBuf, ushort minRunLength)
{
    RLE::iBuf = iBuf;
    RLE::oBuf = oBuf;
    RLE::minRunLength = minRunLength;
}


void
RLE::encodeBlock()
{
    int   i, c;
    word  minCharsLeft = minRunLength + 1;

    while ((iBuf->nCharsLeft() > 0) && (oBuf->nCharsLeft() >= minCharsLeft))
    {
        c = iBuf->getChar();
        runLength = 1;
        while ((iBuf->peekChar() == c) && (runLength < MAX_RUN_LENGTH))
        {
            iBuf->getChar();
            runLength++;
        }
        if (runLength >= minRunLength)
        {
            for (i = 0; i < minRunLength; i++)
            {
                oBuf->putChar(c);
            }
            runLength -= minRunLength;
            oBuf->putChar(runLength);
        }
        else
        {
            for (i = 0; i < runLength; i++)
            {
                oBuf->putChar(c);
            }
        }
    }
    while ((iBuf->nCharsLeft() > 0) && (oBuf->nCharsLeft() > 0))
    {
        oBuf->putChar(iBuf->getChar());
    }
}


void
RLE::decodeBlock()
{
    int  i;

    last = -1;
    runLength = 1;
    while ((iBuf->nCharsLeft() > 0) && (oBuf->nCharsLeft() > 0))
    {
        int c = iBuf->getChar();
        oBuf->putChar(c);
        if (c == last)
        {
            runLength++;
            if ((runLength == minRunLength) && (iBuf->nCharsLeft() > 0))
            {
                runLength = iBuf->getChar();
                for (i = 0; i < runLength; i++)
                {
                    oBuf->putChar(last);
                }
                last = -1;
                runLength = 1;
            }
        }
        else
        {
            last = c;
            runLength = 1;
        }
    }
}
